import os
import sys
import glob

def get_Xtzxy(X,X_ref,tzxy0,resc,target=3):
    tzxy = tzxy0
    Npts =0
    for dist_th in np.linspace(resc,target,5):
        XT = X-tzxy
        ds,inds = cKDTree(X_ref).query(XT)
        keep = ds<dist_th
        X_ref_ = X_ref[inds[keep]]
        X_ = X[keep]
        tzxy = np.mean(X_-X_ref_,axis=0)
        #print(tzxy)
        Npts = np.sum(keep)
    return tzxy,Npts

def get_im_from_Xh(Xh,resc=5,pad=10):
    X = np.round(Xh[:,:3]/resc).astype(int)
    Xm = np.min(X,axis=0)
    XM = np.max(X,axis=0)
    keep = X<=(XM-[0,pad,pad])
    keep &= X>=(Xm+[0,pad,pad])
    keep = np.all(keep,-1)
    X = X[keep]
    if False:
        Xm = np.min(X,axis=0)
        X-=Xm
    else:
        Xm=np.array([0,0,0])

    sz = np.max(X,axis=0)
    imf = np.zeros(sz+1,dtype=np.float32)
    imf[tuple(X.T)]=1
    return imf,Xm

def get_best_translation_points(X,X_ref,resc=10,target=1,return_counts=False):

    im,Xm = get_im_from_Xh(X,resc=resc)
    im_ref,Xm_ref = get_im_from_Xh(X_ref,resc=resc)

    from scipy.signal import fftconvolve
    im_cor = fftconvolve(im,im_ref[::-1,::-1,::-1])
    #plt.imshow(np.max(im_cor,0))
    tzxy = np.array(np.unravel_index(np.argmax(im_cor),im_cor.shape))-im_ref.shape+1+Xm-Xm_ref
    tzxy = tzxy*resc
    Npts=0
    tzxy,Npts = get_Xtzxy(X,X_ref,tzxy,resc=resc,target=target)
    if return_counts:
        return tzxy,Npts
    return tzxy

def get_best_translation_pointsV2(fl,fl_ref,save_folder,set_,resc=5):

	obj = get_dapi_features(fl,save_folder,set_)
	obj_ref = get_dapi_features(fl_ref,save_folder,set_)
	tzxyf,tzxy_plus,tzxy_minus,N_plus,N_minus = np.array([0,0,0]),np.array([0,0,0]),np.array([0,0,0]),0,0
	if (len(obj.Xh_plus)>0) and (len(obj.Xh_minus)>0) and  (len(obj_ref.Xh_plus)>0) and (len(obj_ref.Xh_minus)>0):
		X = obj.Xh_plus[:,:3]
		X_ref = obj_ref.Xh_plus[:,:3]
		tzxy_plus,N_plus = get_best_translation_points(X,X_ref,resc=resc,return_counts=True)

		X = obj.Xh_minus[:,:3]
		X_ref = obj_ref.Xh_minus[:,:3]
		tzxy_minus,N_minus = get_best_translation_points(X,X_ref,resc=resc,return_counts=True)

		tzxyf = -(tzxy_plus+tzxy_minus)/2 

def compute_drift_V2(save_folder,fov,all_flds,set_,redo=False,gpu=True):
	drift_fl = save_folder+os.sep+'driftNew_'+fov.split('.')[0]+'--'+set_+'.pkl'
	if not os.path.exists(drift_fl) or redo:
		fls = [fld+os.sep+fov for fld in all_flds]
		fl_ref = fls[len(fls)//2]
		newdrifts = []
		for fl in fls:
			drft = get_best_translation_pointsV2(fl,fl_ref,save_folder,set_,resc=5)
			newdrifts.append(drft)
		pickle.dump([newdrifts,all_flds,fov,fl_ref],open(drift_fl,'wb'))

if __name__ == "__main__":
	sf = '../MERFISH_Analysis_AER'
	fov = 'Conv_zscan1_001'
	set_ = 'set1'
	all_flds = glob.glob('../MERFISH_Analysis_AER/Conv_zscan1_001--H*_AER_set1--dapiFeatures.npz')
	print(all_flds)
	compute_drift_V2(sf, fov,all_flds, set_)





import subprocess
import sys
import warnings
from pathlib import Path

def _check_cuda():
    message = (
        "\n" + "=" * 80 + "\n"
        + "=" * 80 + "\n"
        + "=" * 80 + "\n"
        "WARNING: This package uses CuPy, which requires the CUDA Toolkit.\n"
        "Make sure you have installed the correct version of CUDA for your GPU.\n"
        "See: https://docs.cupy.dev/en/stable/install.html\n"
        + "=" * 80 + "\n"
        + "=" * 80 + "\n"
        + "=" * 80
    )

    try:
        # Using nvvcc to force an error
        result = subprocess.run(['nvcc', '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if result.returncode != 0:
            raise Exception("NVCC check failed")
    except Exception:
        print(message, file=sys.stderr)
        sys.exit(1)  # Exit installation with error

# Run the check at import time
_check_cuda()

def _get_version():
    # Try to find VERSION file relative to this module
    version_file = Path(__file__).parent / "VERSION"
    if version_file.exists():
        return version_file.read_text().strip()

    # Fall back to looking in parent directory (for development)
    version_file = Path(__file__).parent.parent / "VERSION"
    if version_file.exists():
        return version_file.read_text().strip()

    return "unknown"

__version__ = _get_version()



import os
import sys
import glob
import argparse
import contextlib
from time import sleep,time
from typing import Generator
# Try to import the appropriate TOML library
if sys.version_info >= (3, 11):
	import tomllib  # Python 3.11+ standard library
else:
	import tomli as tomllib  # Backport for older Python versions
import concurrent.futures

# put this first to make sure to capture the correct gpu
os.environ["CUDA_VISIBLE_DEVICES"] = "0"  # Change "1" to the desired GPU ID
import cupy as cp
#cp.cuda.Device(0).use() # The above export doesnt always work so force CuPy to use GPU 0
import numpy as np

#sys.path.pop(0)
from mermake.deconvolver import Deconvolver
from mermake.maxima import find_local_maxima
#from more.maxima import find_local_maxima
from mermake.io import image_generator, load_flats
from mermake.io import ImageQueue, read_cim, dict_to_namespace
import mermake.blur as blur


# Validator for the TOML file
def is_valid_file(path):
    if not os.path.exists(path):
        raise argparse.ArgumentTypeError(f"{path} does not exist.")
    try:
        with open(path, "rb") as f:
            return tomllib.load(f)  # Return raw dict
    except Exception as e:
        raise argparse.ArgumentTypeError(f"Error loading TOML file {path}: {e}")

toml_text = """
        [paths]
        codebook = 'codebook_code_color2__ExtraAaron_8_6_blank.csv' ###
        psf_file = 'dic_psf_60X_cy5_Scope5.pkl'  ### Scope5 psf
        flat_field_tag = 'Scope5_'
        hyb_range = 'H1_AER_set1:H16_AER_set1'
        hyb_folders = [
                        '/data/07_22_2024__PFF_PTBP1',
                        ]
        output_folder = 'MERFISH_Analysis_AER'
        redo = false
        
        #---------------------------------------------------------------------------------------#
        #---------------------------------------------------------------------------------------#
        #           you probably dont have to change any of the settings below                  #
        #---------------------------------------------------------------------------------------#
        #---------------------------------------------------------------------------------------#
        hyb_save = '{fov}--{tag}--col{icol}.npz'
        dapi_save = '{fov}--{tag}--dapiFeatures.npz'
        regex = '''([A-z]+)(\d+)_([^_]+)_set(\d+)(.*)''' #use triple quotes to avoid double escape
        [hybs]
        tile_size = 300
        overlap = 89
        beta = 0.0001
        threshold = 3600
        blur_radius = 30
        delta = 1
        delta_fit = 3
        sigmaZ = 1
        sigmaXY = 1.5
        
        [dapi]
        tile_size = 300
        overlap = 89
        beta = 0.01
        threshold = 3.0
        blur_radius = 50
        delta = 5
        delta_fit = 5
        sigmaZ = 1
        sigmaXY = 1.5"""

class CustomArgumentParser(argparse.ArgumentParser):
	def error(self, message):
		# Customizing the error message
		if "the following arguments are required: settings" in message:
			message = message.replace("settings", "settings.toml")
		message += '\n'
		message += 'The format for the toml file is shown below'
		message += '\n'
		message += toml_text
		super().error(message)

def view_napari(queue, deconvolver, args ):
	image = next(queue)
	buffer = deconvolver.buffer
	flats = deconvolver.flats
	import napari
	viewer = napari.Viewer()
	color = ['red','green','blue', 'white']
	ncol = queue.shape[0]
	for icol in range(ncol-1):
		chan = image[icol]
		flat = flats[icol]
		deconvolver.hybs.apply(chan, flat_field=flat, output=buffer, blur_radius=None)
		deco = cp.asnumpy(buffer)
		deconvolver.hybs.apply(chan, flat_field=flat, output=buffer, **vars(args.dapi))
		Xh = find_local_maxima(buffer, raw = chan, **vars(args.hybs))
		norm = cp.asnumpy(buffer)
		# Stack 3D images: original, deco, norm
		stacked = np.stack([cp.asnumpy(chan), deco, norm], axis=0)
		viewer.add_image(stacked, name=f"channel {icol}", colormap=color[icol], blending='additive')

		# Add corresponding points for this stack
		points = cp.asnumpy(Xh[:, :3])
		viewer.add_points(points, size=7, border_color=color[icol],face_color='transparent', opacity=0.6, name=f"maxima {icol}")
	icol += 1
	chan = image[icol]
	flat = flats[icol]
	deconvolver.dapi.apply(chan, flat_field=flat, output=buffer, blur_radius=None)
	deco = cp.asnumpy(buffer)
	deconvolver.dapi.apply(chan, flat_field=flat, output=buffer, **vars(args.dapi))
	std_val = float(cp.asnumpy(cp.linalg.norm(buffer.ravel()) / cp.sqrt(buffer.size)))
	cp.divide(buffer, std_val, out=buffer)
	Xh_plus = find_local_maxima(buffer, raw = chan, **vars(args.dapi) )
	norm = cp.asnumpy(buffer)
	# Stack 3D images: original, deco, norm
	stacked = np.stack([cp.asnumpy(chan), deco, norm], axis=0)
	viewer.add_image(stacked, name="dapi",  colormap=color[icol], blending='additive')
	points = cp.asnumpy(Xh_plus[:, :3])
	viewer.add_points(points, size=11, border_color=color[icol], face_color='transparent', opacity=0.6, name=f"maxima dapi")
	napari.run()
	exit()


def print_clean(msg, last_len=[0]):
	# Clear previous line
	sys.stdout.write('\r' + ' ' * last_len[0] + '\r')
	sys.stdout.write(msg)
	sys.stdout.flush()
	last_len[0] = len(msg)  # Save new message length


def main():
	prog = sys.argv[0] if sys.argv[0] else "mermake"
	usage = f'{prog} [-opt1, [-opt2, ...]] settings.toml'
	#parser = argparse.ArgumentParser(description='', formatter_class=argparse.RawTextHelpFormatter, usage=usage)
	parser = CustomArgumentParser(description='',formatter_class=argparse.RawTextHelpFormatter,usage=usage)
	parser.add_argument('settings', type=is_valid_file, help='settings file')
	parser.add_argument('-c', '--check', action="store_true", help="Check a single zarr")
	args = parser.parse_args()
	# Convert settings to namespace and attach each top-level section to args
	for key, value in vars(dict_to_namespace(args.settings)).items():
		setattr(args, key, value)
	#----------------------------------------------------------------------------#
	#----------------------------------------------------------------------------#
	psfs = np.load(args.paths.psf_file, allow_pickle=True)

	# the save file executor to do the saving in parallel with computations
	executor = concurrent.futures.ThreadPoolExecutor(max_workers=4)

	
	message = 'Finding input image files.'
	print_clean(message)
	with ImageQueue(args) as queue:
		# set some things based on input images
		ncol = queue.shape[0]
		zpad = queue.shape[1] - 1 # this needs to be about the same size as the input z depth
		sx = queue.shape[2]
		sy = queue.shape[3]
		# this is a buffer to use for copying into 
		buffer = cp.empty(queue.shape[1:], dtype=cp.float32)	
	
		flats = load_flats(shape = queue.shape, **vars(args.paths))

		message = 'Loading PSFs into GPU ram.'
		print_clean(message)
		# these can be very large objects in gpu ram, adjust accoringly to suit gpu specs
		deconvolver = lambda : None
		deconvolver.hybs = Deconvolver(psfs, queue.shape, zpad = zpad, **vars(args.hybs) )
		# shrink the zpad to limit the loaded psfs in ram since dapi isnt deconvolved as strongly
		# or you could just use a single psf, ie (0,1500,1500)
		deconvolver.dapi = Deconvolver(psfs, queue.shape, zpad = zpad//2, **vars(args.dapi))

		if args.check:
			deconvolver.buffer = buffer
			deconvolver.flats = flats
			view_napari(queue, deconvolver, args)

		overlap = args.hybs.overlap
		tile_size = args.hybs.tile_size

		message = 'Starting image processing.\n'
		print_clean(message)
		for image in queue:
			print(image.path, flush=True)
			for icol in range(ncol - 1):
				# there is probably a better way to do the Xh stacking
				Xhf = list()
				chan = image[icol]
				flat = flats[icol]
				for x,y,tile,raw in deconvolver.hybs.tile_wise(chan, flat, **vars(args.hybs)):
					Xh = find_local_maxima(tile, raw = raw, **vars(args.hybs))
					#keep = cp.all((Xh[:,1:3] >= overlap) & (Xh[:,1:3] < tile_size + overlap), axis=-1)
					keep = cp.all((Xh[:,1:3] >= overlap) & (Xh[:,1:3] < cp.array([tile.shape[1] - overlap, tile.shape[2] - overlap])), axis=-1)
					Xh = Xh[keep]
					Xh[:,1] += x - overlap
					Xh[:,2] += y - overlap
					# one more subset to get rid of xfits in the padded area beyond the original image size
					keep = cp.all((Xh[:,1:3] >= 0) & (Xh[:,1:3] < cp.array([sx, sy])), axis=-1)
					Xh = Xh[keep]
					Xhf.append(Xh)
				queue.save_hyb( image.path, icol, Xhf)
				executor.submit(queue.save_hyb, image.path, icol, Xhf)
				del chan, Xhf, Xh


			# now do dapi
			chan = image[-1]
			flat = flats[-1]
			# Deconvolve in-place into the buffer
			deconvolver.dapi.apply(chan, flat_field=flat, output=buffer, **vars(args.dapi))
			# the dapi channel is further normalized by the stdev
			std_val = float(cp.asnumpy(cp.linalg.norm(buffer.ravel()) / cp.sqrt(buffer.size)))
			cp.divide(buffer, std_val, out=buffer)
			Xh_plus = find_local_maxima(buffer, raw = chan, **vars(args.dapi) )
			cp.multiply(buffer, -1, out=buffer)
			Xh_minus = find_local_maxima(buffer, raw = chan, **vars(args.dapi) )
			# save the data
			queue.save_dapi(image.path, icol, Xh_plus, Xh_minus)
			executor.submit(queue.save_dapi, image.path, icol, Xh_plus, Xh_minus)
			image.clear()
			del chan, Xh_plus, Xh_minus, image

if __name__ == "__main__":
	main()



import time,sys
import os,sys,numpy as np

def get_intersV2(self,nmin_bits=3,dinstance_th=2,enforce_color=True,enforce_set=None,redo=False):
	"""Get an initial intersection of points and save in self.res"""
	self.res_fl = self.decoded_fl.replace('decoded','res')
	if not os.path.exists(self.res_fl) or redo:
		res =[]
		if enforce_color and (enforce_set is None):
			icols = self.XH[:,-2].astype(int)
			XH = self.XH
			for icol in tqdm(np.unique(icols)):
				inds = np.where(icols==icol)[0]
				Xs = XH[inds,:3]
				Ts = cKDTree(Xs)
				res_ = Ts.query_ball_tree(Ts,dinstance_th)
				res += [inds[r] for r in res_]
		elif enforce_color and (enforce_set is not None):
			ibits = self.XH[:,-1].astype(int)
			isets = ibits//enforce_set
			icols = self.XH[:,-2].astype(int)
			XH = self.XH
			for icol in np.unique(icols):
				for iset in tqdm(np.unique(isets)):
					inds = np.where((icols==icol)&(isets==iset))[0]
					Xs = XH[inds,:3]
					Ts = cKDTree(Xs)
					res_ = Ts.query_ball_tree(Ts,dinstance_th)
					res += [inds[r] for r in res_]
		else:
			XH = self.XH
			Xs = XH[:,:3]
			Ts = cKDTree(Xs)
			res = Ts.query_ball_tree(Ts,dinstance_th)
		print("Calculating lengths of clusters...")
		lens = np.array(list(map(len,res)))
		Mlen = np.max(lens)
		print("Unfolding indexes...")
		res_unfolder = np.concatenate(res)
		print("Saving to file:",self.res_fl)
		self.res_unfolder=res_unfolder
		self.lens=lens
		
		#np.savez(self.res_fl,res_unfolder=res_unfolder,lens=lens)
	else:
		dic = np.load(self.res_fl)
		self.res_unfolder=dic['res_unfolder']
		self.lens=dic['lens']
		#self.res = res
	lens =self.lens
	self.res_unfolder = self.res_unfolder[np.repeat(lens, lens)>=nmin_bits]
	self.lens = self.lens[lens>=nmin_bits]
def get_icodesV3(dec,nmin_bits=3,iH=-3):
	import time
	start = time.time()
	lens = dec.lens
	res_unfolder = dec.res_unfolder
	Mlen = np.max(lens)
	print("Calculating indexes within cluster...")
	res_is = np.tile(np.arange(Mlen), len(lens))
	res_is = res_is[res_is < np.repeat(lens, Mlen)]
	print("Calculating index of molecule...")
	ires = np.repeat(np.arange(len(lens)), lens)
	#r0 = np.array([r[0] for r in res for r_ in r])
	print("Calculating index of first molecule...")
	r0i = np.concatenate([[0],np.cumsum(lens)])[:-1]
	r0 = res_unfolder[np.repeat(r0i, lens)]
	print("Total time unfolded molecules:",time.time()-start)
	
	### torch
	ires = torch.from_numpy(ires.astype(np.int64))
	res_unfolder = torch.from_numpy(res_unfolder.astype(np.int64))
	res_is = torch.from_numpy(res_is.astype(np.int64))
	
	import time
	start = time.time()
	print("Computing score...")
	scoreF = torch.from_numpy(dec.XH[:,iH])[res_unfolder]
	print("Total time computing score:",time.time()-start)
	
	
	### organize molecules in blocks for each cluster
	def get_asort_scores():
		val = torch.max(scoreF)+2
		scoreClu = torch.zeros([len(lens),Mlen],dtype=torch.float64)+val
		scoreClu[ires,res_is]=scoreF
		asort = scoreClu.argsort(-1)
		scoreClu = torch.gather(scoreClu,dim=-1,index=asort)
		scoresF2 = scoreClu[scoreClu<val-1]
		return asort,scoresF2
	def get_reorder(x,val=-1):
		if type(x) is not torch.Tensor:
			x = torch.from_numpy(np.array(x))
		xClu = torch.zeros([len(lens),Mlen],dtype=x.dtype)+val
		xClu[ires,res_is] = x
		xClu = torch.gather(xClu,dim=-1,index=asort)
		xf = xClu[xClu>val]
		return xf
	
	
	import time
	start = time.time()
	print("Computing sorting...")
	asort,scoresF2 = get_asort_scores()
	res_unfolder2 = get_reorder(res_unfolder,val=-1)
	del asort
	del scoreF
	print("Total time sorting molecules by score:",time.time()-start)
	
	import time
	start = time.time()
	print("Finding best bits per molecules...")
	
	Rs = dec.XH[:,-1].astype(np.int64)
	Rs = torch.from_numpy(Rs)
	Rs_U = Rs[res_unfolder2]
	nregs,nbits = dec.codes_01.shape
	score_bits = torch.zeros([len(lens),nbits],dtype=scoresF2.dtype)-1
	score_bits[ires,Rs_U]=scoresF2
	
	
	codes_lib = torch.from_numpy(np.array(dec.codes__))
	
	
	codes_lib_01 = torch.zeros([len(codes_lib),nbits],dtype=score_bits.dtype)
	for icd,cd in enumerate(codes_lib):
		codes_lib_01[icd,cd]=1
	codes_lib_01 = codes_lib_01/torch.norm(codes_lib_01,dim=-1)[:,np.newaxis]
	print("Finding best code...")
	batch = 10000
	icodes_best = torch.zeros(len(score_bits),dtype=torch.int64)
	dists_best = torch.zeros(len(score_bits),dtype=torch.float32)
	from tqdm import tqdm
	for i in tqdm(range((len(score_bits)//batch)+1)):
		score_bits_ = score_bits[i*batch:(i+1)*batch]
		if len(score_bits_)>0:
			score_bits__ = score_bits_.clone()
			score_bits__[score_bits__==-1]=0
			score_bits__ = score_bits__/torch.norm(score_bits__,dim=-1)[:,np.newaxis]
			Mul = torch.matmul(score_bits__,codes_lib_01.T)
			max_ = torch.max(Mul,dim=-1)
			icodes_best[i*batch:(i+1)*batch] = max_.indices
			dists_best[i*batch:(i+1)*batch] = 2-2*max_.values
	
	
	keep_all_bits = torch.sum(score_bits.gather(1,codes_lib[icodes_best])>=0,-1)>=nmin_bits
	dists_best_ = dists_best[keep_all_bits]
	score_bits = score_bits[keep_all_bits]
	icodes_best_ = icodes_best[keep_all_bits]
	icodesN=icodes_best_
	
	indexMols_ = torch.zeros([len(lens),nbits],dtype=res_unfolder2.dtype)-1
	indexMols_[ires,Rs_U]=res_unfolder2
	indexMols_ = indexMols_[keep_all_bits]
	indexMols_ = indexMols_.gather(1,codes_lib[icodes_best_])
	
	# make unique
	indexMols_,rinvMols = get_unique_ordered(indexMols_)
	icodesN = icodesN[rinvMols]
	
	XH = torch.from_numpy(dec.XH)
	XH_pruned = XH[indexMols_]
	XH_pruned[indexMols_==-1]=np.nan
	
	dec.dist_best = dists_best_[rinvMols].numpy()
	dec.XH_pruned=XH_pruned.numpy()
	dec.icodesN=icodesN.numpy()
	np.savez_compressed(dec.decoded_fl,XH_pruned=dec.XH_pruned,icodesN=dec.icodesN,gns_names = np.array(dec.gns_names),dist_best=dec.dist_best)
	print("Total time best bits per molecule:",time.time()-start)

def compute_decoding(save_folder,fov,set_,lib_fl, redo=False):
	dec = decoder_simple(save_folder,fov,set_)
	complete = dec.check_is_complete()
	if complete==0 or redo:
		#compute_drift(save_folder,fov,all_flds,set_,redo=False,gpu=False)
		dec = decoder_simple(save_folder,fov=fov,set_=set_)
		dec.get_XH(fov,set_,ncols=3,nbits=100,th_h=3600,tag_keep='_AER_')#number of colors match 
		dec.XH = dec.XH[dec.XH[:,-4]>0.25] ### keep the spots that are correlated with the expected PSF for 60X
		dec.load_library(lib_fl,nblanks=-1)
		
		dec.ncols = 3
		get_intersV2(dec,nmin_bits=3,dinstance_th=2,enforce_color=True,enforce_set=None,redo=False)
		get_icodesV3(dec,nmin_bits=3,iH=-3)
		#dec.get_inters(dinstance_th=2,enforce_color=True)# enforce_color=False
		#dec.get_inters(dinstance_th=2,nmin_bits=4,enforce_color=True,redo=True)
		#dec.get_icodes(nmin_bits=4,method = 'top4',norm_brightness=None,nbits=24)#,is_unique=False)
		#get_icodesV2(dec,nmin_bits=4,delta_bits=None,iH=-3,redo=False,norm_brightness=False,nbits=24,is_unique=True)



import os
import re
import gc
import glob
from fnmatch import fnmatch
from types import SimpleNamespace
from typing import List, Tuple, Optional
import json
import argparse
from argparse import Namespace

import xml.etree.ElementTree as ET
import zarr
import cupy as cp
import numpy as np

import concurrent.futures

from . import blur
from . import __version__

def center_crop(A, shape):
	"""Crop numpy array to (h, w) from center."""
	h, w = shape[-2:]
	H, W = A.shape
	top = (H - h) // 2
	left = (W - w) // 2
	return A[top:top+h, left:left+w]

def load_flats(flat_field_tag, shape=None, **kwargs):
	stack = list()
	files = sorted(glob.glob(flat_field_tag + '*'))
	for file in files:
		im = np.load(file)['im']
		if shape is not None:
			im = center_crop(im, shape)
		cim = cp.array(im,dtype=cp.float32)
		blurred = blur.box(cim, (20,20))
		blurred = blurred / cp.median(blurred)
		stack.append(blurred)
	return cp.stack(stack)

def get_iH(fld): return int(os.path.basename(fld).split('_')[0][1:])
def get_files(master_data_folders, set_ifov,iHm=None,iHM=None):
	#if not os.path.exists(save_folder): os.makedirs(save_folder)
	all_flds = []
	for master_folder in master_data_folders:
		all_flds += glob.glob(master_folder+os.sep+r'H*_AER_*')
		#all_flds += glob.glob(master_folder+os.sep+r'H*_Igfbpl1_Aldh1l1_Ptbp1*')
	### reorder based on hybe
	all_flds = np.array(all_flds)[np.argsort([get_iH(fld)for fld in all_flds])] 
	set_,ifov = set_ifov
	all_flds = [fld for fld in all_flds if set_ in os.path.basename(fld)]
	all_flds = [fld for fld in all_flds if ((get_iH(fld)>=iHm) and (get_iH(fld)<=iHM))]
	#fovs_fl = save_folder+os.sep+'fovs__'+set_+'.npy'
	folder_map_fovs = all_flds[0]#[fld for fld in all_flds if 'low' not in os.path.basename(fld)][0]
	fls = glob.glob(folder_map_fovs+os.sep+'*.zarr')
	fovs = np.sort([os.path.basename(fl) for fl in fls])
	fov = fovs[ifov]
	all_flds = [fld for fld in all_flds if os.path.exists(fld+os.sep+fov)]
	return all_flds,fov

def read_im(path, return_pos=False):
	dirname = os.path.dirname(path)
	fov = os.path.basename(path).split('_')[-1].split('.')[0]
	file_ = os.path.join(dirname, fov, 'data')

	z = zarr.open(file_, mode='r')
	image = np.array(z[1:])

	shape = image.shape
	xml_file = os.path.splitext(path)[0] + '.xml'
	if os.path.exists(xml_file):
		txt = open(xml_file, 'r').read()
		tag = '<z_offsets type="string">'
		zstack = txt.split(tag)[-1].split('</')[0]

		tag = '<stage_position type="custom">'
		x, y = eval(txt.split(tag)[-1].split('</')[0])

		nchannels = int(zstack.split(':')[-1])
		nzs = (shape[0] // nchannels) * nchannels
		image = image[:nzs].reshape([shape[0] // nchannels, nchannels, shape[-2], shape[-1]])
		image = image.swapaxes(0, 1)

	if image.dtype == np.uint8:
		image = image.astype(np.uint16) ** 2

	if return_pos:
		return image, x, y
	return image

class Container:
	def __init__(self, data, **kwargs):
		# Store the array and any additional metadata
		self.data = data
		self.metadata = kwargs
	def __getitem__(self, item):
		# Allow indexing into the container
		return self.data[item]
	def __array__(self):
		# Return the underlying array
		return self.data
	def __repr__(self):
		# Custom string representation showing the metadata or basic info
		return f"Container(shape={self.data.shape}, dtype={self.data.dtype}, metadata={self.metadata})"
	def __getattr__(self, name):
		# If attribute is not found on the container, delegate to the CuPy object
		if hasattr(self.data, name):
			return getattr(self.data, name)
		raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")
	def clear(self):
		# Explicitly delete the CuPy array and synchronize
		if hasattr(self, 'data') and self.data is not None:
			del self.data
			self.data = None

def read_cim(path):
	im = read_im(path)
	cim = cp.asarray(im)
	container = Container(cim)
	container.path = path
	return container

def read_ccim(path, return_pos=False):
	dirname = os.path.dirname(path)
	fov = os.path.basename(path).split('_')[-1].split('.')[0]
	file_ = os.path.join(dirname, fov, 'data')

	z = zarr.open(file_, mode='r')
	nz = z.shape[0]

	# Skip z[0], start at z[1]
	slices = []
	for i in range(1, nz):
		slices.append(cp.asarray(z[i]))

	image = cp.stack(slices)
	shape = image.shape

	xml_file = os.path.splitext(path)[0] + '.xml'
	if os.path.exists(xml_file):
		txt = open(xml_file, 'r').read()
		tag = '<z_offsets type="string">'
		zstack = txt.split(tag)[-1].split('</')[0]

		tag = '<stage_position type="custom">'
		x, y = eval(txt.split(tag)[-1].split('</')[0])

		nchannels = int(zstack.split(':')[-1])
		nzs = (shape[0] // nchannels) * nchannels
		image = image[:nzs].reshape((shape[0] // nchannels, nchannels, shape[-2], shape[-1]))
		image = cp.swapaxes(image, 0, 1)

	if image.dtype == cp.uint8:
		image = image.astype(cp.uint16) ** 2

	container = Container(image)
	container.path = path

	if return_pos:
		return container, x, y
	return container

def get_ifov(zarr_file_path):
	"""Extract ifov from filename - finds last digits before .zarr"""
	filename = Path(zarr_file_path).name  # Keep full filename with extension
	match = re.search(r'([0-9]+)[^0-9]*\.zarr', filename)
	if match:
		return int(match.group(1))
	raise ValueError(f"No digits found before .zarr in filename: {filename}")

class FolderFilter:
	def __init__(self, hyb_range: str, regex_pattern: str):
		self.hyb_range = hyb_range
		self.regex = re.compile(regex_pattern)
		self.start_pattern, self.end_pattern = self.hyb_range.split(':')
		
		# Parse start and end patterns
		self.start_parts = self._parse_pattern(self.start_pattern)
		self.end_parts = self._parse_pattern(self.end_pattern)
		
	def _parse_pattern(self, pattern: str) -> Optional[Tuple]:
		"""Parse a pattern using the regex to extract components"""
		match = self.regex.match(pattern)
		if match:
			return match.groups()
		return None
	
	def _extract_numeric_part(self, text: str) -> int:
		"""Extract numeric part from text like 'H1' -> 1"""
		match = re.search(r'\d+', text)
		return int(match.group()) if match else 0
	
	def _compare_patterns(self, file_parts: Tuple, start_parts: Tuple, end_parts: Tuple) -> bool:
		"""
		Compare if file_parts falls within the range defined by start_parts and end_parts
		Groups: (prefix, number, middle, set_number, suffix)
		"""
		if not all([file_parts, start_parts, end_parts]):
			return False
			
		# Extract components
		file_prefix, file_num, file_middle, file_set, file_suffix = file_parts
		start_prefix, start_num, start_middle, start_set, start_suffix = start_parts
		end_prefix, end_num, end_middle, end_set, end_suffix = end_parts
		
		# Convert to integers for comparison
		file_num = int(file_num)
		file_set = int(file_set)
		start_num = int(start_num)
		start_set = int(start_set)
		end_num = int(end_num)
		end_set = int(end_set)
	
		# Check if middle part matches (e.g., 'MER')
		if start_middle == '*':
			pass
		elif file_middle != start_middle or file_middle != end_middle:
			return False
			
		# Check if prefix matches
		if file_prefix != start_prefix or file_prefix != end_prefix:
			return False
			
		num_in_range = start_num <= file_num <= end_num
		set_in_range = start_set <= file_set <= end_set
		
		return num_in_range and set_in_range
	
	def isin(self, text: str) -> bool:
		"""Check if a single text/filename falls within the specified range"""
		file_parts = self._parse_pattern(text)
		if not file_parts:
			return False
		return self._compare_patterns(file_parts, self.start_parts, self.end_parts)
	
	def filter_files(self, filenames: List[str]) -> List[str]:
		"""Filter filenames that fall within the specified range"""
		matching_files = []
		
		for filename in filenames:
			if self.isin(filename):
				matching_files.append(filename)
				
		return matching_files

class ImageQueue:
	__version__ = __version__
	def __init__(self, args):
		self.args = args

		self.args_array = namespace_to_array(self.args.settings)
		self.__dict__.update(vars(args.paths))

		os.makedirs(self.output_folder, exist_ok = True)
		
		self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)

		folder_filter = FolderFilter(self.hyb_range, self.regex)
		
		fov_min, fov_max = (-float('inf'), float('inf'))
		if hasattr(self, "fov_range"):
			fov_min, fov_max = map(int, self.fov_range.split(':'))
	
		matches = list()
		for root in self.hyb_folders:
			if not os.path.exists(root):
				continue
			try:
				with os.scandir(root) as entries:
					for sub in entries:
						if sub.is_dir(follow_symlinks=False) and folder_filter.isin(sub.name):
							try:
								with os.scandir(sub.path) as items:
									for item in items:
										if item.is_dir(follow_symlinks=False) and '.zarr' in item.name:
											ifov = get_ifov(str(item.name))
											if fov_min <= ifov <= fov_max:
												matches.append(item.path)
							except PermissionError:
								continue
			except PermissionError:
				continue
		
		self.files = iter(sorted(matches))

		# Preload the first valid image
		self._first_image = self._load_first_valid_image()
		self.shape = self._first_image.shape
		self.dtype = self._first_image.dtype
		
		# Only redo analysis if it is true
		if hasattr(self, "redo") and not self.redo:
			# Filter out already processed files
			filtered = [f for f in sorted(matches) if not self._is_done(f)]
			# Reload first valid image from sorted list
			self.files = iter(filtered)
			self._first_image = self._load_first_valid_image()

		# Start prefetching the next image
		self.future = None
		self._prefetch_next_image()

	def _load_first_valid_image(self):
		"""Try loading images one by one until one succeeds."""
		for file in self.files:
			try:
				future = self.executor.submit(read_cim, file)
				image = future.result()
				return image
			except Exception as e:
				#print(f"Warning: Failed to load image {file}: {e}")
				continue
		raise RuntimeError("No valid images could be found.")

	def _prefetch_next_image(self):
		try:
			next_file = next(self.files)
			self.future = self.executor.submit(read_cim, next_file)
		except StopIteration:
			self.future = None

	def __iter__(self):
		return self

	def __next__(self):
		if self._first_image is not None:
			image = self._first_image
			self._first_image = None
			return image
		while self.future is not None:
			try:
				image = self.future.result()
			except Exception as e:
				#print(f"Warning: Failed to load image: {e}")
				image = None
			# Try to prefetch the next image regardless
			try:
				next_file = next(self.files)
				self.future = self.executor.submit(read_cim, next_file)
			except StopIteration:
				self.future = None
			# If we got a valid image, return it
			if image is not None:
				return image

		# If we reach here, there are no more images in the current batch
		if False:
			# In watch mode, look for new files
			import time
			time.sleep(60)
			
			# Find any new files
			new_matches = self._find_matching_files()
			# Filter to only files we haven't processed yet
			new_matches = [f for f in new_matches if f not in self.processed_files]
			
			if new_matches:
				# New files found!
				new_matches.sort()
				self.matches = new_matches
				self.files = iter(self.matches)
				self.processed_files.update(new_matches)  # Mark as seen
				
				# Prefetch the first new image
				self._prefetch_next_image()
				
				# Try again to get the next image
				return self.__next__()
			else:
				# No new files yet, but we'll keep watching
				return self.__next__()

		self.close()
		raise StopIteration

	def __enter__(self):
		return self

	def __exit__(self, exc_type, exc_val, exc_tb):
		self.close()

	def _is_done(self, path):
		fov, tag = self.path_parts(path)
		for icol in range(self.shape[0] - 1):
			filename = self.hyb_save.format(fov=fov, tag=tag, icol=icol)
			filepath = os.path.join(self.output_folder, filename)
			if not os.path.exists(filepath):
				return False
		filename = self.dapi_save.format(fov=fov, tag=tag, icol=icol)
		filepath = os.path.join(self.output_folder, filename)
		if not os.path.exists(filepath):
			return False
		return True


	def close(self):
		self.executor.shutdown(wait=True)

	def path_parts(self, path):
		path_obj = Path(path)
		fov = path_obj.stem  # The filename without extension
		tag = path_obj.parent.name  # The parent directory name (which you seem to want)
		return fov, tag

	def save_hyb(self, path, icol, Xhf, attempt=1, max_attempts=3):
		fov,tag = self.path_parts(path)
		filename = self.hyb_save.format(fov=fov, tag=tag, icol=icol)
		filepath = os.path.join(self.output_folder, filename)
	
		Xhf = [x for x in Xhf if x.shape[0] > 0]
		if Xhf:
			xp = cp.get_array_module(Xhf[0])
			Xhf = xp.vstack(Xhf)
		else:
			xp = np
			Xhf = np.array([])
		if not os.path.exists(filepath) or (hasattr(self, "redo") and self.redo):
			xp.savez_compressed(filepath, Xh=Xhf, version=__version__, args=self.args_array)
			#  Optional integrity check after saving
			# this seems to greatly slow everything down
			#try:
			#	with np.load(filepath) as dat:
			#		_ = dat["Xh"].shape  # Try accessing a key
			#except Exception as e:
			#	os.remove(filepath)
			#	if attempt < max_attempts:
			#		return self.save_hyb(path, icol, Xhf, attempt=attempt+1, max_attempts=max_attempts)
			#	else:
			#		raise RuntimeError(f"Failed saving xfit file after {max_attempts} attempts: {filepath}")
		del Xhf
		if xp == cp:
			xp._default_memory_pool.free_all_blocks()  # Free standard GPU memory pool

	def save_dapi(self, path, icol, Xh_plus, Xh_minus, attempt=1, max_attempts=3):
		fov, tag = self.path_parts(path)
		filename = self.dapi_save.format(fov=fov, tag=tag, icol=icol)
		filepath = os.path.join(self.output_folder, filename)
	
		xp = cp.get_array_module(Xh_plus)
		if not os.path.exists(filepath) or (hasattr(self, "redo") and self.redo):
			xp.savez_compressed(filepath, Xh_plus=Xh_plus, Xh_minus=Xh_minus, version=__version__, args=self.args_array)
			#  Optional integrity check after saving
			# this seems to greatly slow everything down
			#try:
			#	with np.load(filepath) as dat:
			#		_ = dat["Xh_minus"].shape  # Try accessing a key
			#except Exception as e:
			#	os.remove(filepath)
			#	if attempt < max_attempts:
			#		return self.save_dapi(path, icol, Xh_plus, Xh_minus, attempt=attempt+1, max_attempts=max_attempts)
			#	else:
			#		raise RuntimeError(f"Failed saving xfit file after {max_attempts} attempts: {filepath}")
		del Xh_plus, Xh_minus
		if xp == cp:
			xp._default_memory_pool.free_all_blocks()


def image_generator(hybs, fovs):
	"""Generator that prefetches the next image while processing the current one."""
	with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
		future = None
		for all_flds, fov in zip(hybs, fovs):
			for hyb in all_flds:

				file = os.path.join(hyb, fov)
				next_future = executor.submit(read_cim, file)
				if future:
					yield future.result()
				future = next_future
		if future:
			yield future.result()


from .utils import *
def read_xml(path):
	# Open and parse the XML file
	tree = None
	with open(path, "r", encoding="ISO-8859-1") as f:
		tree = ET.parse(f)
	return tree.getroot()

def get_xml_field(file, field):
	xml = read_xml(file)
	return xml.find(f".//{field}").text
def set_data(args):
	from wcmatch import glob as wc
	from natsort import natsorted
	pattern = args.paths.hyb_range
	batch = dict()
	files = list()
	# parse hybrid folders
	files = find_files(**vars(args.paths))
	for file in files:
		sset = re.search('_set[0-9]*', file).group()
		hyb = os.path.basename(os.path.dirname(file))
		#hyb = re.search(pattern, file).group()
		if sset and hyb:
			batch.setdefault(sset, {}).setdefault(os.path.basename(file), {})[hyb] = {'zarr' : file}
	# parse xml files
	points = list()
	for sset in sorted(batch):
		for fov in sorted(batch[sset]):
			point = list()
			for hyb,dic in natsorted(batch[sset][fov].items()):
				path = dic['zarr']
				#file = glob.glob(os.path.join(dirname,'*' + basename + '.xml'))[0]
				file = path.replace('zarr','xml')
				point.append(list(map(float, get_xml_field(file, 'stage_position').split(','))))
			mean = np.mean(np.array(point), axis=0)
			batch[sset][fov]['stage_position'] = mean
			points.append(mean)
	points = np.array(points)
	mins = np.min(points, axis=0)
	step = estimate_step_size(points)
	#coords = points_to_coords(points)
	for sset in sorted(batch):
		for i,fov in enumerate(sorted(batch[sset])):
			point = batch[sset][fov]['stage_position']
			point -= mins
			batch[sset][fov]['grid_position'] = np.round(point / step).astype(int)
	args.batch = batch
	#counts = Counter(re.search(pattern, file).group().split('_set')[0] for file in files if re.search(pattern, file))
	#hybrid_count = {key: counts[key] for key in natsorted(counts)}

def dict_to_namespace(d):
	"""Recursively convert dictionary into SimpleNamespace."""
	for key, value in d.items():
		if isinstance(value, dict):
			value = dict_to_namespace(value)
		elif isinstance(value, list):
			value = [dict_to_namespace(i) if isinstance(i, dict) else i for i in value]
		d[key] = value
	return SimpleNamespace(**d)
def namespace_to_dict(obj):
	"""Recursively convert namespace objects to dictionaries"""
	if isinstance(obj, argparse.Namespace):
		return {k: namespace_to_dict(v) for k, v in vars(obj).items()}
	elif isinstance(obj, list):
		return [namespace_to_dict(item) for item in obj]
	elif isinstance(obj, dict):
		return {k: namespace_to_dict(v) for k, v in obj.items()}
	else:
		return obj

def namespace_to_array(obj, prefix=''):
	"""
	Recursively convert Namespace or dict to list of (block, key, value) tuples.
	prefix is the accumulated parent keys joined by dots.
	"""
	rows = []
	if isinstance(obj, (Namespace, SimpleNamespace)):
		obj = vars(obj)
	if isinstance(obj, dict):
		for k, v in obj.items():
			full_key = f"{prefix}.{k}" if prefix else k
			if isinstance(v, (Namespace, SimpleNamespace, dict)):
				rows.extend(namespace_to_array(v, prefix=full_key))
			else:
				rows.append((prefix, k, str(v)))
	else:
		# For other types just append
		rows.append((prefix, '', str(obj)))
	return rows



import os
import cupy as cp

# Get path relative to this file's location
this_dir = os.path.dirname(__file__)
cu_path = os.path.join(this_dir, "blur.cu")
with open(cu_path, "r") as f:
	kernel_code = f.read()

# Define the kernels
box_1d_kernel = cp.RawKernel(kernel_code, "box_1d")
#optimized_box_1d_kernel = cp.RawKernel(kernel_code, "optimized_box_1d")
#box_plane_kernel = cp.RawKernel(kernel_code, "box_plane")

def box(image, size, axes=None, output=None, temp=None):
	"""
	Apply separable box blur on 2D or 3D cupy arrays.

	Parameters
	----------
	image : cupy.ndarray
		Input 2D or 3D array (will not be modified).
	size : int or tuple of int
		Blur size per axis. If int, the same size is used for all specified axes.
	axes : tuple of int, optional
		Axes to blur. If None, blur all axes.
		For 2D: (0, 1) means blur both dimensions
		For 3D: (0, 1, 2) means blur all dimensions
		Examples: (0,) for axis 0 only, (1, 2) for axes 1 and 2
	temp : cupy.ndarray, optional
		Temporary buffer for multi-axis blurs. If None, one is created.
		Must be same shape and dtype as input.
	temp : cupy.ndarray, optional
		Temporary buffer. If None, one is created.
	output : cupy.ndarray, optional
		Output array. If None, a new one is created.

	Returns
	-------
	cupy.ndarray
		Blurred array.
	"""
	# The kernel expects float32
	if image.dtype != cp.float32:
		image = image.astype(cp.float32)

	# Input validation
	if image.ndim not in (2, 3):
		raise ValueError("Only 2D or 3D arrays are supported")

	# Handle axes parameter
	if axes is None:
		axes = tuple(range(image.ndim))  # All axes
	elif isinstance(axes, int):
		axes = (axes,)  # Single axis
	elif not isinstance(axes, (tuple, list)):
		raise TypeError("axes must be None, int, tuple, or list")

	# Validate and normalize axes (handle negative indices)
	normalized_axes = []
	for axis in axes:
		if axis < 0:
			axis = image.ndim + axis
		if axis < 0 or axis >= image.ndim:
			raise ValueError(f"axis {axis} is out of bounds for array of dimension {image.ndim}")
		normalized_axes.append(axis)

	# Remove duplicates and sort
	axes = tuple(sorted(set(normalized_axes)))

	if len(axes) == 0:
		# No axes to blur, just copy
		if output is None:
			return cp.copy(image)
		else:
			cp.copyto(output, image)
			return output

	# Handle size parameter
	if isinstance(size, int):
		sizes = {axis: size for axis in axes}
	elif isinstance(size, (tuple, list)):
		if len(size) != len(axes):
			raise ValueError(f"Size must have {len(axes)} elements for {len(axes)} axes")
		sizes = dict(zip(axes, size))
	else:
		raise TypeError("size must be an int, tuple, or list of ints")

	# Create output if necessary
	if output is None:
		output = cp.empty_like(image)

	# If only one axis, do single blur
	if len(axes) == 1:
		axis = axes[0]
		box_1d(image, sizes[axis], axis=axis, output=output)
		return output

	# For multiple axes, we need exactly one temporary buffer
	# (can't avoid this due to CUDA kernel race condition constraints)
	
	# WE CAN DO WITHOUT THE EXTRA TEMP BUFFER IF WE DO A TRUE MULTIDIMENSIONAL
	# NESTED FOR LOOPS IN THE CUDA KERNEL INSTEAD OF MULTIPLE LINEAR SEPARABLE 1D
	# THE TRADEOFF IS SPEED VERSUS RAM


	if temp is None:
		temp = cp.empty_like(image)
	else:
		# Validate provided temp buffer
		if temp.shape != image.shape or temp.dtype != image.dtype:
			raise ValueError("temp buffer must have same shape and dtype as input")

	# First blur: image -> output
	box_1d(image, sizes[axes[0]], axis=axes[0], output=output)

	# Remaining blurs alternate between output and temp
	for i in range(1, len(axes)):
		axis = axes[i]
		if i % 2 == 1:  # Odd iterations: output -> temp
			box_1d(output, sizes[axis], axis=axis, output=temp)
		else:  # Even iterations: temp -> output
			box_1d(temp, sizes[axis], axis=axis, output=output)

	# If we ended on temp, copy back to output
	if len(axes) % 2 == 0:  # Even number of axes means we ended on temp
		cp.copyto(output, temp)

	return output


def box_2d(image, size, axes=(-2, -1), output=None):
	"""
	Convenience function for 2D blur on specified axes.

	Parameters
	----------
	image : cupy.ndarray
		Input array (2D or 3D).
	size : int or tuple of 2 ints
		Blur size for the two axes.
	axes : tuple of 2 ints, default (-2, -1)
		Which two axes to blur.
	output : cupy.ndarray, optional
		Output array. If None, a new one is created.

	Returns
	-------
	cupy.ndarray
		Blurred array.
	"""
	if len(axes) != 2:
		raise ValueError("axes must contain exactly 2 elements for 2D blur")

	return box(image, size, axes=axes, output=output)

def box_1d(image, size, axis=0, output=None):
	if image.dtype != cp.float32:
		image = image.astype(cp.float32)
	if output is None:
		output = cp.empty_like(image)

	# Prevent in-place operations that cause race conditions
	assert output is not image, "In-place operation not supported - input and output must be different"

	if axis < 0:
		axis = image.ndim - 1

	delta = size // 2
	
	if image.ndim == 2:
		# For 2D arrays, assume XY format
		size_x, size_y = image.shape
		size_z = 1  # Dummy dimension
		assert axis < 2, f"axis {axis} is out of bounds for array of dimension 2"
		axis += 1
	elif image.ndim == 3:
		# For 3D arrays, use ZXY format
		size_z, size_x, size_y = image.shape
	else:
		raise ValueError("Only 2D or 3D arrays are supported")
	
	threads_per_block = 256
	blocks = (size_z * size_x * size_y + threads_per_block - 1) // threads_per_block
	box_1d_kernel((blocks,), (threads_per_block,),
				  (image, output, size_z, size_x, size_y, delta, axis))
	return output



